//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"io/ioutil"
	"log"
	"os"
)

type ProgressCB func(step int, stepName string, current, total float64, userdata interface{}) bool
type DoneCB func(result Result, userdata interface{}) bool

type ProgressData struct {
	Step     int
	StepName string
	Current  float64
	Total    float64
}

type Result struct {
	ResponseCode int
	ErrorString  string
}

type AttachmentChunk struct {
	Data  []byte
	Error error
}

type ImportContext struct {
	conf                 *Config
	stdlog               *log.Logger
	dbglog               *log.Logger
	GroupName            string
	WorkDir              string
	AttachmentChan       chan AttachmentChunk
	ProgressCallBack     ProgressCB
	ProgressCallBackData interface{}
	Cancel               <-chan bool

	// TODO: add source config
}

func NewImportContext(conf *Config, stdlog, dbglog *log.Logger) *ImportContext {
	if stdlog == nil {
		stdlog = log.New(ioutil.Discard, "", 0)
	}
	if dbglog == nil {
		dbglog = log.New(ioutil.Discard, "", 0)
	}

	ctx := &ImportContext{}
	ctx.conf = conf
	ctx.stdlog = stdlog
	ctx.dbglog = dbglog
	ctx.GroupName = ""
	ctx.AttachmentChan = make(chan AttachmentChunk, 32)
	ctx.ProgressCallBack = nil
	ctx.Cancel = nil
	return ctx
}

func (ctx *ImportContext) createTempWorkDir() (err error) {
	ctx.WorkDir, err = ioutil.TempDir(ctx.conf.TempPath, "tank-")
	return
}

func (ctx *ImportContext) removeTempWorkDir() {
	if err := os.RemoveAll(ctx.WorkDir); err != nil {
		ctx.stdlog.Printf("Error removing WorkDir: %s", err)
	}
	return
}

func (ctx *ImportContext) reportProgress(step int, stepName string, current, total float64) {
	if ctx.ProgressCallBack != nil {
		if keep := ctx.ProgressCallBack(step, stepName, current, total, ctx.ProgressCallBackData); !keep {
			ctx.ProgressCallBack = nil
		}
	}
}

func (ctx *ImportContext) isCanceled() bool {
	return ctx.Cancel != nil && len(ctx.Cancel) > 0
}
