//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"crypto/rand"
	"encoding/base64"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"strings"
	"time"
)

type newSessionResponse struct {
	id           string
	session      *Session
	responsecode int
	errorstring  string
}

type newSessionRequest struct {
	ctx      *ImportContext
	refId    string
	response chan newSessionResponse
}

type getSessionResponse struct {
	session      *Session
	refId        string
	responsecode int
	errorstring  string
}

type getSessionRequest struct {
	group    string
	id       string
	response chan getSessionResponse
}

type SessionsUpdateCB func(added, removed map[string]string, userdata interface{}) bool

type SessionsListCB struct {
	cb       SessionsUpdateCB
	userdata interface{}
}

type listSessionsResponse struct {
	sessions     map[string]string
	responsecode int
	errorstring  string
}

type listSessionsRequest struct {
	group    string
	callback SessionsUpdateCB
	userdata interface{}
	response chan listSessionsResponse
}

type removeSessionResponse struct {
	responsecode int
	errorstring  string
}

type removeSessionRequest struct {
	group    string
	id       string
	response chan removeSessionResponse
}

type sessionStoreSessionElement struct {
	s     *session
	refId string
}

type sessionStoreGroupElement struct {
	sessions  map[string]*sessionStoreSessionElement
	updateCBs []SessionsListCB
}

func (group *sessionStoreGroupElement) callUpdateHandler(added, removed map[string]string) {
	var keptCBs []SessionsListCB
	for _, cb := range group.updateCBs {
		if cb.cb != nil {
			if keep := cb.cb(added, removed, cb.userdata); keep {
				keptCBs = append(keptCBs, cb)
			}
		}
	}
	group.updateCBs = keptCBs
}

func (group *sessionStoreGroupElement) callUpdateHandlerAdd(id, refId string) {
	added := make(map[string]string)
	added[id] = refId
	group.callUpdateHandler(added, nil)
}

func (group *sessionStoreGroupElement) callUpdateHandlerRemove(id, refId string) {
	removed := make(map[string]string)
	removed[id] = refId
	group.callUpdateHandler(nil, removed)
}

type sessionStore struct {
	store      map[string]*sessionStoreGroupElement
	conf       *Config
	stdlog     *log.Logger
	dbglog     *log.Logger
	quit       chan bool
	done       chan bool
	newChan    chan newSessionRequest
	getChan    chan getSessionRequest
	listChan   chan listSessionsRequest
	removeChan chan removeSessionRequest
}

func generateSessionId() (string, error) {
	var b [32]byte
	if _, err := rand.Read(b[:]); err != nil {
		return "", err
	}
	return base64.RawURLEncoding.EncodeToString(b[:]), nil
}

func (store *sessionStore) new(ctx *ImportContext, refId string) (resp newSessionResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"
	id, err := generateSessionId()
	if err != nil {
		resp.responsecode = http.StatusInternalServerError
		resp.errorstring = err.Error()
		return
	}

	resp.id = id
	if _, exists := store.store[ctx.GroupName]; !exists {
		newgroup := &sessionStoreGroupElement{}
		newgroup.sessions = make(map[string]*sessionStoreSessionElement)
		store.store[ctx.GroupName] = newgroup
	}
	ctx.conf = store.conf
	if pref := ctx.stdlog.Prefix(); strings.Contains(pref, "%s") {
		ctx.stdlog.SetPrefix(fmt.Sprintf(pref, resp.id))
	}
	if pref := ctx.dbglog.Prefix(); strings.Contains(pref, "%s") {
		ctx.dbglog.SetPrefix(fmt.Sprintf(pref, resp.id))
	}

	s, err := newSession(ctx, func() { store.GetInterface().Remove(ctx.GroupName, resp.id) })
	if err != nil {
		resp.responsecode = http.StatusInternalServerError
		resp.errorstring = err.Error()
		return
	}
	store.store[ctx.GroupName].sessions[resp.id] = &sessionStoreSessionElement{s, refId}
	resp.session = store.store[ctx.GroupName].sessions[resp.id].s.getInterface()
	store.dbglog.Printf("SessionStore: created session for '%s' -> %s", ctx.GroupName, resp.id)
	store.store[ctx.GroupName].callUpdateHandlerAdd(resp.id, refId)
	return
}

func (store *sessionStore) get(groupname, id string) (resp getSessionResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"

	group, exists := store.store[groupname]
	if !exists {
		resp.responsecode = http.StatusNotFound
		resp.errorstring = fmt.Sprintf("SessionStore: session '%s/%s' not found", groupname, id)
		return
	}

	if session, exists := group.sessions[id]; exists {
		resp.session = session.s.getInterface()
		resp.refId = session.refId
		return
	}
	resp.responsecode = http.StatusNotFound
	resp.errorstring = fmt.Sprintf("SessionStore: session '%s/%s' not found", groupname, id)
	return
}

func (store *sessionStore) list(groupname string, userdata interface{}, cb SessionsUpdateCB) (resp listSessionsResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"
	resp.sessions = make(map[string]string)
	if group, exists := store.store[groupname]; exists {
		for id, e := range group.sessions {
			resp.sessions[id] = e.refId
		}
		if cb != nil {
			group.updateCBs = append(group.updateCBs, SessionsListCB{cb, userdata})
		}
	} else if cb != nil {
		newgroup := &sessionStoreGroupElement{}
		newgroup.sessions = make(map[string]*sessionStoreSessionElement)
		newgroup.updateCBs = []SessionsListCB{SessionsListCB{cb, userdata}}
		store.store[groupname] = newgroup
	}
	return
}

func (store *sessionStore) remove(groupname, id string) (resp removeSessionResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"

	group, exists := store.store[groupname]
	if !exists {
		resp.responsecode = http.StatusNotFound
		resp.errorstring = fmt.Sprintf("SessionStore: session '%s/%s' not found", groupname, id)
		return
	}

	if session, exists := group.sessions[id]; exists {
		go session.s.cleanup() // cleanup could take a while -> don't block all the other stuff
		refId := session.refId
		delete(group.sessions, id)
		store.dbglog.Printf("SessionStore: removed session '%s/%s'", groupname, id)
		group.callUpdateHandlerRemove(id, refId)

		if len(group.sessions) == 0 && len(group.updateCBs) == 0 {
			delete(store.store, groupname)
			store.dbglog.Printf("SessionStore: removed group '%s'", groupname)
		}
	} else {
		resp.responsecode = http.StatusNotFound
		resp.errorstring = fmt.Sprintf("SessionStore: session '%s/%s' not found", groupname, id)
	}
	return
}

func (store *sessionStore) maintenanceTask() {
	for name, group := range store.store {
		group.callUpdateHandler(nil, nil)
		if len(group.sessions) == 0 && len(group.updateCBs) == 0 {
			delete(store.store, name)
			store.dbglog.Printf("SessionStore: removed group '%s'", name)
		}
	}
}

func (store *sessionStore) dispatchRequests() {
	defer func() { store.done <- true }()

	mt := time.NewTicker(1 * time.Minute)
	for {
		select {
		case <-store.quit:
			return
		case <-mt.C:
			store.maintenanceTask()
		case req := <-store.newChan:
			req.response <- store.new(req.ctx, req.refId)
		case req := <-store.getChan:
			req.response <- store.get(req.group, req.id)
		case req := <-store.listChan:
			req.response <- store.list(req.group, req.userdata, req.callback)
		case req := <-store.removeChan:
			req.response <- store.remove(req.group, req.id)
		}
	}
}

// *********************************************************
// Public Interface

type SessionStore struct {
	newChan    chan<- newSessionRequest
	getChan    chan<- getSessionRequest
	listChan   chan listSessionsRequest
	removeChan chan<- removeSessionRequest
}

func (store *SessionStore) New(ctx *ImportContext, refId string) (string, *Session, int, string) {
	resCh := make(chan newSessionResponse)
	req := newSessionRequest{}
	req.ctx = ctx
	req.refId = refId
	req.response = resCh
	store.newChan <- req

	res := <-resCh
	return res.id, res.session, res.responsecode, res.errorstring
}

func (store *SessionStore) Get(group, id string) (*Session, string, int, string) {
	resCh := make(chan getSessionResponse)
	req := getSessionRequest{}
	req.group = group
	req.id = id
	req.response = resCh
	store.getChan <- req

	res := <-resCh
	return res.session, res.refId, res.responsecode, res.errorstring
}

func (store *SessionStore) List(group string, userdata interface{}, cb SessionsUpdateCB) (map[string]string, int, string) {
	resCh := make(chan listSessionsResponse)
	req := listSessionsRequest{}
	req.group = group
	req.response = resCh
	req.callback = cb
	req.userdata = userdata
	store.listChan <- req

	res := <-resCh
	return res.sessions, res.responsecode, res.errorstring
}

func (store *SessionStore) Remove(group, id string) (int, string) {
	resCh := make(chan removeSessionResponse)
	req := removeSessionRequest{}
	req.group = group
	req.id = id
	req.response = resCh
	store.removeChan <- req

	res := <-resCh
	return res.responsecode, res.errorstring
}

func (store *sessionStore) GetInterface() *SessionStore {
	ch := &SessionStore{}
	ch.newChan = store.newChan
	ch.getChan = store.getChan
	ch.listChan = store.listChan
	ch.removeChan = store.removeChan
	return ch
}

func (store *sessionStore) Cleanup() {
	store.quit <- true
	<-store.done
	close(store.quit)
	close(store.done)
	close(store.newChan)
	close(store.getChan)
	close(store.listChan)
	close(store.removeChan)
}

func NewSessionStore(conf *Config, stdlog, dbglog *log.Logger) (store *sessionStore, err error) {
	if stdlog == nil {
		stdlog = log.New(ioutil.Discard, "", 0)
	}
	if dbglog == nil {
		dbglog = log.New(ioutil.Discard, "", 0)
	}

	store = &sessionStore{}
	store.conf = conf
	store.stdlog = stdlog
	store.dbglog = dbglog
	store.quit = make(chan bool, 1)
	store.done = make(chan bool)
	store.store = make(map[string]*sessionStoreGroupElement)
	store.newChan = make(chan newSessionRequest, 10)
	store.getChan = make(chan getSessionRequest, 10)
	store.listChan = make(chan listSessionsRequest, 10)
	store.removeChan = make(chan removeSessionRequest, 10)

	go store.dispatchRequests()
	return
}
