//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package store

import (
	"fmt"
	"io"
	"os"
	"os/user"
	"path/filepath"
	"testing"
)

var (
	testBasePath = "run/aura-tank_testing"
	testDBPath   = "/run/aura-tank_testing/db.bolt"
)

func TestMain(m *testing.M) {
	u, err := user.Current()
	if err != nil {
		os.Exit(-1)
	}
	testBasePath = fmt.Sprintf("/run/user/%s/aura-tank_testing", u.Uid)
	testDBPath = filepath.Join(testBasePath, ".db.bolt")
	os.Exit(m.Run())
}

//
// Testing
//

func TestOpen(t *testing.T) {
	// base-path is non-existing directory
	if _, err := NewStore(Config{"/nonexistend/"}); err == nil {
		t.Fatalf("opening store in nonexisting directory should throw an error")
	}

	// store db-path is a directory
	os.Remove(testDBPath)
	if err := os.MkdirAll(testDBPath, 0700); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if _, err := NewStore(Config{testBasePath}); err == nil {
		t.Fatalf("opening store using a directory should throw an error")
	}

	// exisiting but non-database file
	os.Remove(testDBPath)
	if f, err := os.Create(testDBPath); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		io.WriteString(f, "this is not a bolt db.")
		f.Close()
	}
	if _, err := NewStore(Config{testBasePath}); err == nil {
		t.Fatalf("opening store using a invalid database should throw an error")
	}

	// TODO: add checks for
	//    - missing/wrong version entry
	//    - missing/wrong buckets (ie. store intregrity)

	// create new bolt-db and reopen it
	os.Remove(testDBPath)
	store, err := NewStore(Config{testBasePath})
	if err != nil {
		t.Fatalf("creating new store failed: %v", err)
	}
	store.Close()

	store, err = NewStore(Config{testBasePath})
	if err != nil {
		t.Fatalf("re-opening existing store failed: %v", err)
	}
	if _, err := NewStore(Config{testBasePath}); err == nil {
		t.Fatalf("opening already opened database should throw an error")
	}
	store.Close()
}
